# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import uuid
from typing import List, Tuple

import pytest

from uim.codec.parser.base import SupportedFormats
from uim.model import UUIDIdentifier
from uim.model.base import InkModelException
from uim.model.ink import InkModel, InkTree
from uim.model.inkdata.brush import VectorBrush, BrushPolygon, BrushPolygonUri, RasterBrush, RotationMode, BlendMode
from uim.model.inkdata.strokes import Spline, Style, Stroke, LayoutMask, PathPointProperties
from uim.model.inkinput.inputdata import Environment, InkInputProvider, InkInputType, InputDevice, SensorChannel, \
    InkSensorType, InkSensorMetricType, SensorChannelsContext, SensorContext, InputContext, DataType, \
    InputContextRepository
from uim.model.inkinput.sensordata import SensorData, InkState
from uim.model.semantics import schema
from uim.model.semantics.node import StrokeGroupNode, StrokeNode, StrokeFragment
from uim.model.semantics.schema import SemanticTriple
from uim.utils.matrix import Matrix4x4


def pen_sensor_context(ink_model: InkModel, provider: InkInputProvider) -> Tuple[SensorChannelsContext,
                                                                                 List[SensorChannel]]:
    """
    Create a sensor context for a pen.

    Parameters
    ----------
    ink_model: InkModel
        The ink model.
    provider: InkInputProvider
        The input provider.

    Returns
    -------
    Tuple[SensorChannelsContext, List[SensorChannel]]
        Sensor channels context and the list of sensor channels.
    """
    input_device_pen: InputDevice = InputDevice()
    input_device_pen.properties.append(("dev.id", "345456567"))
    input_device_pen.properties.append(("dev.manufacturer", "Wacom"))
    ink_model.input_configuration.devices.append(input_device_pen)
    # Sensor channels
    sensor_channels_bluetooth: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.FORCE, resolution=1.0)
    ]
    # Create a sensor channel context
    scc_bluetooth: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                                 ink_input_provider_id=provider.id,
                                                                 channels=sensor_channels_bluetooth)
    assert input_device_pen != provider
    return scc_bluetooth, sensor_channels_bluetooth


def create_ink_model() -> InkModel:
    # Create the model
    ink_model: InkModel = InkModel(version=SupportedFormats.UIM_VERSION_3_1_0.value)
    # Setting a unit scale factor
    ink_model.unit_scale_factor = 1.5
    # Using a 4x4 matrix for scaling
    ink_model.transform = Matrix4x4.create_scale(1.5)
    # Create an environment
    env: Environment = Environment()
    env.add_environment_property("env.id", uuid.uuid4().hex)
    assert str(env)
    assert env != ink_model
    ink_model.input_configuration.environments.append(env)
    # Ink input provider can be pen, mouse or touch.
    provider: InkInputProvider = InkInputProvider(input_type=InkInputType.MOUSE)
    provider.properties.append(("dev.id", "1234567"))
    assert provider != env
    ink_model.input_configuration.ink_input_providers.append(provider)
    # Input device is the sensor (pen tablet, screen, etc.)
    input_device: InputDevice = InputDevice()
    input_device.properties.append(("dev.manufacturer", "Wacom"))
    ink_model.input_configuration.devices.append(input_device)
    # Create a group of sensor channels
    sensor_channels_tablet: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.FORCE, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.AZIMUTH, metric=InkSensorMetricType.ANGLE, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.ALTITUDE, metric=InkSensorMetricType.ANGLE, resolution=1.0)
    ]
    # Create a sensor channel context
    scc_tablet: SensorChannelsContext = SensorChannelsContext(channels=sensor_channels_tablet,
                                                              ink_input_provider_id=provider.id,
                                                              input_device_id=input_device.id)
    # We can create an additional input device, for example one providing pressure via Bluetooth
    scc_bluetooth, _ = pen_sensor_context(ink_model, provider)
    # Add all sensor channel contexts
    sensor_context: SensorContext = SensorContext()
    sensor_context.add_sensor_channels_context(scc_tablet)
    sensor_context.add_sensor_channels_context(scc_bluetooth)
    ink_model.input_configuration.sensor_contexts.append(sensor_context)
    # Create the input context using the Environment and the Sensor Context
    input_context: InputContext = InputContext(environment_id=env.id, sensor_context_id=sensor_context.id)
    ink_model.input_configuration.input_contexts.append(input_context)
    sensor_data_id: uuid.UUID = UUIDIdentifier.id_generator()
    # Create sensor data
    sensor_data_0: SensorData = SensorData(sensor_data_id, input_context_id=input_context.id,
                                           state=InkState.PLANE)
    sensor_data_0.add_timestamp_data(sensor_channels_tablet[2], [0., 0.1, 0.2, 0.3, 0.4])
    sensor_data_0.add_data(sensor_channels_tablet[0], [100.4, 103.7, 110.1])
    sensor_data_0.add_data(sensor_channels_tablet[1], [200.1, 202.0, 207.0])
    sensor_data_0.add_data(sensor_channels_tablet[3], [0.1, 0.4, 0.2])
    sensor_data_0.add_data(sensor_channels_tablet[4], [50., 51.0, 52.0])
    sensor_data_0.add_data(sensor_channels_tablet[5], [50., 51.0, 52.0])

    # Add sensor data to the model
    ink_model.sensor_data.add(sensor_data_0)
    # We need to define a brush polygon
    points: list = [(10, 10), (0, 10), (0, 0)]
    brush_polygons: list = [BrushPolygon(min_scale=0., points=points)]
    # PathPointProperties
    path_point_properties = PathPointProperties(
        size=1.0,
        red=0.0,
        green=1.0,
        blue=0.0,
        alpha=1.0,
        rotation=0.0,
        scale_x=1.0,
        scale_y=1.0,
        scale_z=1.0,
        offset_x=0.0,
        offset_y=0.0,
        offset_z=0.0
    )
    # Create the brush object using polygons
    vector_brush_0: VectorBrush = VectorBrush("app://qa-test-app/vector-brush/MyTriangleBrush", brush_polygons)
    # Add it to the model
    ink_model.brushes.add_vector_brush(vector_brush_0)
    # Add a brush specified with shape Uris
    poly_uris: list = [
        BrushPolygonUri("will://brush/3.0/shape/Circle?precision=20&radius=1", 0.),
        BrushPolygonUri("will://brush/3.0/shape/Ellipse?precision=20&radiusX=1&radiusY=0.5", 4.0)
    ]
    vector_brush_1: VectorBrush = VectorBrush(
        "app://qa-test-app/vector-brush/MyEllipticBrush",
        poly_uris)
    raster_brush_0: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush",
        spacing=10., scattering=5., rotation=RotationMode.TRAJECTORY, shape_textures=[bytes([10, 20]),
                                                                                      bytes([30, 20])],
        fill_width=2.0, fill_height=0.3,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        randomize_fill=False, blend_mode=BlendMode.SOURCE_OVER)
    assert input_context != raster_brush_0
    # Add it to the model
    ink_model.brushes.add_raster_brush(raster_brush_0)
    raster_brush_1: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2., fill_height=0.3,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=False, blend_mode=BlendMode.SOURCE_OVER)
    ink_model.brushes.add_raster_brush(raster_brush_1)
    ink_model.brushes.add_vector_brush(vector_brush_1)
    # Specify the layout of the stroke data, in this case the stroke will have variable X, Y and Size properties.
    layout_mask: int = LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value
    # Provide the stroke data - in this case 4 data points, each consisting of X, Y, Size
    path: list = [10., 10.7, 1.0, 21.0, 20.2, 2.0, 30.0, 12.4, 2.1, 40.0, 15.0, 1.5]
    # Create a spline object from the path data
    spline: Spline = Spline(layout_mask, path)
    # Create some style
    style: Style = Style(brush_uri=vector_brush_0.name)
    style.path_point_properties.red = 1.0
    style.path_point_properties.green = 0.0
    style.path_point_properties.blue = 0.0
    style.path_point_properties.alpha = 1.0
    # Create a stroke object. Note that it just exists, but is not in the model yet.
    stroke_0: Stroke = Stroke(sid=UUIDIdentifier.id_generator(), spline=spline, style=style)
    # Create a spline object - 9 data points, each consisting of X, Y, Size, Red, Green, Blue, Alpha
    spline_1: Spline = Spline(
        LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value | LayoutMask.RED.value | LayoutMask.GREEN.value
        | LayoutMask.BLUE.value | LayoutMask.ALPHA.value,
        [10.0, 10.7, 1.0, 0.5, 0.0, 0.1, 1.0,
         21.0, 20.2, 2.0, 0.9, 0.4, 0.2, 0.8,
         30.0, 12.4, 2.1, 0.7, 0.1, 0.1, 0.7,
         40.0, 15.0, 1.5, 0.3, 0.5, 0.4, 1.0,
         50.0, 45.0, 1.0, 0.3, 0.5, 0.4, 1.0,
         41.0, 53.0, 1.1, 0.2, 0.3, 0.5, 0.9,
         33.0, 73.0, 1.2, 0.6, 0.7, 0.4, 0.8,
         20.0, 84.0, 1.3, 0.7, 0.8, 0.3, 0.7,
         10.0, 91.0, 1.1, 0.7, 0.9, 0.2, 0.6])
    spline_2: Spline = Spline(
        LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value,
        [20., 20., 1.0, 21.0, 20.2, 2.0, 30.0, 12.4, 2.1, 40.0, 15.0, 1.5])
    # Create a style
    style_1: Style = Style(brush_uri=raster_brush_0.name)
    style_1.path_point_properties.rotation = 0.35
    # Create another style
    style_2: Style = Style(properties=path_point_properties,
                           brush_uri=vector_brush_1.name)

    # The render mode URI can also be app specific like app://blabla
    # The URI will://rasterization/3.0/blend-mode/SourceOver is assumed and must not be set.
    style_1.render_mode_uri = "will://rasterization/3.0/blend-mode/DestinationOver"
    # Create a stroke object. Note that it just exists, but is not in the model yet.
    stroke_1: Stroke = Stroke(UUIDIdentifier.id_generator(), spline=spline_1, style=style_1)
    stroke_2: Stroke = Stroke(UUIDIdentifier.id_generator(), spline=spline_2, style=style_2)
    # First you need a root group to contain the strokes
    root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    main_tree: InkTree = InkTree()
    ink_model.add_tree(main_tree)
    # Assign the group as the root of the main ink tree
    ink_model.ink_tree.root = root
    # Add a node for stroke 0
    stroke_node_0: StrokeNode = StrokeNode(stroke_0, StrokeFragment(0, 1, 0.0, 1.0))
    root.add(stroke_node_0)
    ink_model.knowledge_graph.append(SemanticTriple(stroke_node_0.uri, schema.IS, "will:ml/0.2/MathStroke"))
    # Add a node for stroke 1
    root.add(StrokeNode(stroke_1, StrokeFragment(0, 1, 0.0, 1.0)))
    # Add third node
    root.add(StrokeNode(stroke_2))
    return ink_model


def test_sensor_channel():
    """
    Test the sensor channel structure.
    """
    channel_id: uuid.UUID = UUIDIdentifier.id_generator()
    input_provider_id: uuid.UUID = UUIDIdentifier.id_generator()
    input_device_id: uuid.UUID = UUIDIdentifier.id_generator()
    test_sensor_channel_1: SensorChannel = SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                         metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                         channel_min=0., channel_max=10., precision=2, index=0,
                                                         name="X Channel", data_type=DataType.FLOAT32,
                                                         ink_input_provider_id=input_provider_id,
                                                         input_device_id=input_device_id)
    assert test_sensor_channel_1 == SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert str(test_sensor_channel_1)
    assert test_sensor_channel_1 == SensorChannel(channel_id=UUIDIdentifier.id_generator(),
                                                  channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.Y,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.TIME, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=10.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=5., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=15., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=4, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X-Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.INT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=UUIDIdentifier.id_generator(),
                                                  input_device_id=input_device_id)
    assert test_sensor_channel_1 != SensorChannel(channel_id=channel_id, channel_type=InkSensorType.X,
                                                  metric=InkSensorMetricType.LENGTH, resolution=1.,
                                                  channel_min=0., channel_max=10., precision=2, index=0,
                                                  name="X Channel", data_type=DataType.FLOAT32,
                                                  ink_input_provider_id=input_provider_id,
                                                  input_device_id=UUIDIdentifier.id_generator())


def test_input_configuration():
    """Creates an ink model from the scratch."""
    # Create the model
    ink_model: InkModel = create_ink_model()
    # Check the environment
    for env in ink_model.input_configuration.environments:
        assert str(env)
        for prop in env.properties:
            assert str(prop)
            if prop[0] == "env.name":
                assert prop[1] == "My Environment"
            if prop[0] == "os.id":
                assert prop[1] == "98765"
            if prop[0] == "os.name":
                assert prop[1] == "Windows"
            if prop[0] == "os.version":
                assert prop[1] == "10.0.18362.239"
            if prop[0] == "os.build":
                assert prop[1] == "239"
            if prop[0] == "os.platform":
                assert prop[1] == "whatever"
    # Check the input providers
    for provider in ink_model.input_configuration.ink_input_providers:
        assert str(provider)
        assert provider.type in [InkInputType.PEN, InkInputType.MOUSE]
        for prop in provider.properties:
            assert str(prop)
            if prop[0] == "dev.id":
                assert prop[1] == "1234567"
    # Check the input devices
    for device in ink_model.input_configuration.devices:
        assert str(device)
        for prop in device.properties:
            assert str(prop)
            if prop[0] == "dev.id":
                assert prop[1] in ["123454321", "345456567"]
            if prop[0] == "dev.manufacturer":
                assert prop[1] == "Wacom"
    # Check the sensor contexts
    for sensor_context in ink_model.input_configuration.sensor_contexts:
        assert str(sensor_context)
        assert sensor_context.id is not None
        for scc in sensor_context.sensor_channels_contexts:
            assert str(scc)

            for sc in scc.channels:
                assert str(sc)
    # Check the input contexts
    for input_context in ink_model.input_configuration.input_contexts:
        assert str(input_context)
        assert input_context.environment_id is not None
        assert input_context.sensor_context_id is not None


def test_input_and_sensor_context():
    """
    Test sensor context.
    """
    input_device_pen: InputDevice = InputDevice()
    input_device_pen.properties.append(("dev.id", "345456567"))
    input_device_pen.properties.append(("dev.manufacturer", "Wacom"))
    provider: InkInputProvider = InkInputProvider(input_type=InkInputType.PEN)
    environment: Environment = Environment()
    environment.properties.append(("env.name", "My Environment"))
    # Sensor channels
    sensor_channels: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.FORCE, resolution=1.0)
    ]
    # Create a sensor channel context
    scc: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                       ink_input_provider_id=provider.id,
                                                       channels=sensor_channels)

    # Azimuth channel has not been added
    with pytest.raises(InkModelException):
        scc.get_channel_by_type(InkSensorType.AZIMUTH)
    scc.add_sensor_channel(SensorChannel(channel_type=InkSensorType.AZIMUTH, metric=InkSensorMetricType.ANGLE,
                                         resolution=1.0))
    azimuth_channel = scc.get_channel_by_type(InkSensorType.AZIMUTH)
    assert azimuth_channel is not None

    assert scc != provider
    # Different sensor channels
    sensor_channels_2: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1.0)
    ]
    scc_2: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                         ink_input_provider_id=provider.id,
                                                         channels=sensor_channels_2)
    assert scc != scc_2
    # Different sensor channels
    sensor_channels_mod: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=2.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.FORCE, resolution=1.0)
    ]
    scc_2_b: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                           ink_input_provider_id=provider.id,
                                                           channels=sensor_channels_mod)
    assert scc != scc_2_b
    # Different sampling rate
    scc_3: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                         ink_input_provider_id=provider.id,
                                                         channels=sensor_channels, sampling_rate_hint=512)
    assert scc != scc_3
    # Different latency
    scc_4: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                         ink_input_provider_id=provider.id,
                                                         channels=sensor_channels, latency=100)
    assert scc != scc_4
    # Different input device id
    scc_5: SensorChannelsContext = SensorChannelsContext(input_device_id=UUIDIdentifier.id_generator(),
                                                         ink_input_provider_id=provider.id,
                                                         channels=sensor_channels)
    assert scc != scc_5
    # Different ink input provider
    scc_6: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                         ink_input_provider_id=UUIDIdentifier.id_generator(),
                                                         channels=sensor_channels)
    assert scc != scc_6
    # Add all sensor channel contexts
    sensor_context: SensorContext = SensorContext()
    sensor_context.add_sensor_channels_context(scc)
    with pytest.raises(InkModelException):
        sensor_context.get_channel_by_id(UUIDIdentifier.id_generator())
    with pytest.raises(InkModelException):
        sensor_context.get_channel_by_type(InkSensorType.RADIUS_X)

    sensor_context_2: SensorContext = SensorContext()
    assert sensor_context != sensor_context_2
    sensor_context_2.add_sensor_channels_context(scc_2)
    assert sensor_context != scc
    assert sensor_context != sensor_context_2
    environment_id: uuid.UUID = UUIDIdentifier.id_generator()
    input_context: InputContext = InputContext(environment_id=environment_id,
                                               sensor_context_id=sensor_context.id)
    input_context_2: InputContext = InputContext(environment_id=environment_id,
                                                 sensor_context_id=UUIDIdentifier.id_generator())
    assert input_context != input_context_2
    input_context_repo: InputContextRepository = InputContextRepository()
    input_context_repo.add_sensor_context(sensor_context)
    input_context_repo.add_ink_device(input_device_pen)
    input_context_repo.add_input_context(input_context)
    input_context_repo.add_environment(environment)
    input_context_repo.add_input_provider(provider)
    assert str(input_context_repo)
    # Check if the sensor context can be retrieved
    assert input_context_repo.get_sensor_context(sensor_context.id) == sensor_context
    assert input_context_repo.get_input_device(input_device_pen.id) == input_device_pen
    # Unknown uuids should raise an exception
    with pytest.raises(InkModelException):
        input_context_repo.get_input_context(UUIDIdentifier.id_generator())
    with pytest.raises(InkModelException):
        input_context_repo.get_sensor_context(UUIDIdentifier.id_generator())
    with pytest.raises(InkModelException):
        input_context_repo.get_input_device(UUIDIdentifier.id_generator())
    # Check if the has configuration method works
    assert input_context_repo.has_configuration()
    assert input_context_repo != sensor_context
    # Now compare the repositories
    input_context_repo_2: InputContextRepository = InputContextRepository()
    assert input_context_repo != input_context_repo_2
    input_context_repo_2.add_input_context(input_context)
    assert input_context_repo != input_context_repo_2
    input_context_repo_2.add_sensor_context(sensor_context)
    assert input_context_repo != input_context_repo_2
    input_context_repo_2.add_input_provider(provider)
    assert input_context_repo != input_context_repo_2
    input_context_repo_2.add_ink_device(input_device_pen)
    assert input_context_repo != input_context_repo_2
    input_context_repo_2.add_environment(environment)
    # Now the repositories should be equal
    assert input_context_repo == input_context_repo_2
    # Setup different context: Sensor Context
    input_context_repo_other: InputContextRepository = InputContextRepository()
    input_context_repo_other.add_sensor_context(sensor_context_2)
    input_context_repo_other.add_ink_device(input_device_pen)
    input_context_repo_other.add_input_context(input_context)
    input_context_repo_other.add_environment(environment)
    input_context_repo_other.add_input_provider(provider)
    assert input_context_repo != input_context_repo_other
    # Setup different context: Input device
    input_context_repo_other: InputContextRepository = InputContextRepository()
    input_context_repo_other.add_sensor_context(sensor_context)
    input_context_repo_other.add_ink_device(InputDevice())
    input_context_repo_other.add_input_context(input_context)
    input_context_repo_other.add_environment(environment)
    input_context_repo_other.add_input_provider(provider)
    assert input_context_repo != input_context_repo_other
    # Setup different context: Input context
    input_context_repo_other: InputContextRepository = InputContextRepository()
    input_context_repo_other.add_sensor_context(sensor_context)
    input_context_repo_other.add_ink_device(input_device_pen)
    input_context_repo_other.add_input_context(InputContext())
    input_context_repo_other.add_environment(environment)
    input_context_repo_other.add_input_provider(provider)
    assert input_context_repo != input_context_repo_other
    # Setup different context: Environment
    input_context_repo_other: InputContextRepository = InputContextRepository()
    input_context_repo_other.add_sensor_context(sensor_context)
    input_context_repo_other.add_ink_device(input_device_pen)
    input_context_repo_other.add_input_context(input_context)
    input_context_repo_other.add_environment(Environment())
    input_context_repo_other.add_input_provider(provider)
    assert input_context_repo != input_context_repo_other
    # Setup different context: Input  Provider
    input_context_repo_other: InputContextRepository = InputContextRepository()
    input_context_repo_other.add_sensor_context(sensor_context)
    input_context_repo_other.add_ink_device(input_device_pen)
    input_context_repo_other.add_input_context(input_context)
    input_context_repo_other.add_environment(environment)
    input_context_repo_other.add_input_provider(InkInputProvider())
    assert input_context_repo != input_context_repo_other



# -*- coding: utf-8 -*-
"""
This package contains the Universal Ink Model codec and a decoder for the Wacom Ink Layer Language (WILL) format.
"""
__all__ = ['context', 'parser', 'writer']

from uim.codec import context
from uim.codec import parser
from uim.codec import writer



# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import math
import uuid
from typing import List

import numpy as np
import pytest

from uim.model import UUIDIdentifier
from uim.model.base import Identifier
from uim.model.helpers.boundingbox import union, union_all
from uim.model.helpers.spline import catmull_rom_one_point, catmull_rom, linear_interpol
from uim.model.inkdata.strokes import PathPointProperties
from uim.model.inkinput.inputdata import unit2unit, Unit, unit2unit_matrix
from uim.model.semantics.structures import BoundingBox
from uim.utils.matrix import Matrix4x4


def test_matrix_helper():
    """
    Test the matrix helper functions.
    """
    matrix: List[List[float]] = Matrix4x4.create_scale(1.5)
    assert matrix[0] == [1.5, 0, 0., 0.]
    assert matrix[1] == [0., 1.5, 0., 0.]
    assert matrix[2] == [0., 0, 1.5, 0.]
    assert matrix[3] == [0., 0, 0., 1.]
    matrix = Matrix4x4.create_translation([1.0, 2.0, 3.0])
    assert matrix[0] == [1., 0., 0., 1.]
    assert matrix[1] == [0., 1., 0., 2.]
    assert matrix[2] == [0., 0., 1., 3.]
    assert matrix[3] == [0., 0., 0., 1.]
    try:
        Matrix4x4.create_translation([1.0, 2.0, 3.0, 4.0])
        pytest.fail("Translation must be a 3D vector. Exception should have been thrown.")
    except ValueError:
        pass


def test_spline_helper():
    """
    Test the spline helper functions.
    """
    value = catmull_rom_one_point(0.5, 0.0, 1.0, 2.0, 3.0)
    assert value == 1.5


def test_spline_array():
    """
    Test the spline helper functions.
    """
    spline_x, spline_y = catmull_rom([0.0, 1.0, 2.0, 3.0], [0.0, 1.0, 2.0, 3.0], 2)
    assert spline_x == [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0]
    assert spline_y == [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0]
    spline_x, spline_y = catmull_rom([0.0, 1.0, 2.0], [0.0, 1.0, 2.0], 2)
    assert spline_x == [0.0, 0.5, 1.0, 1.5, 2.0]
    assert spline_y == [0.0, 0.5, 1.0, 1.5, 2.0]
    spline_x, spline_y = catmull_rom([0.0], [0.0], 2)
    assert spline_x == [0.0]
    assert spline_y == [0.0]
    try:
        catmull_rom([], [], 2)
        pytest.fail("Empty arrays should throw an exception.")
    except ValueError:
        print("Empty arrays throw an exception.")


def test_linear_interpolation():
    """
    Test the linear interpolation function.
    """
    lin_x, lin_y = linear_interpol([0.0, 1.0, 2.0], [0.0, 1.0, 2.0])
    assert lin_x == pytest.approx([0.0, 0.6666666, 1.3333333, 2.0])
    assert lin_y == pytest.approx([0.0, 0.6666666, 1.3333333, 2.0])
    lin_x, lin_y = linear_interpol([0.0, 1.0], [0.0, 1.0])
    assert lin_x == pytest.approx([0.0, 0.3333333, 0.6666666, 1.0])
    assert lin_y == pytest.approx([0.0, 0.3333333, 0.6666666, 1.0])
    lin_x, lin_y = linear_interpol([0.0, 20., 1.0], [0.0, 20., 1.0])
    assert lin_x == pytest.approx([0.0, 0.3333333, 0.6666666, 1.0])
    assert lin_y == pytest.approx([0.0, 0.3333333, 0.6666666, 1.0])
    try:
        linear_interpol([0.], [0.])
        pytest.fail("Empty arrays should throw an exception.")
    except ValueError:
        print("Empty arrays throw an exception.")


def test_bounding_box():
    """
    Test the bounding box function.
    """
    bb1 = BoundingBox(x=0.0, y=0.0, width=100., height=100.)
    bb2 = BoundingBox(x=10.0, y=10.0, width=100., height=100.)
    bb3 = union(bb1, bb2)
    assert bb3.x == 0.0
    assert bb3.y == 0.0
    assert bb3.width == 110.0
    assert bb3.height == 110.0
    bb4 = union_all([bb1, bb2, bb3])
    assert bb4.x == 0.0
    assert bb4.y == 0.0
    assert bb4.width == 110.0
    assert bb4.height == 110.0
    assert str(bb4) == "<Bounding box : [x:=0.0, y:=0.0, width:=110.0, height:=110.0]>"


def test_unit_conversion():
    """
    Test the unit conversion function.
    """
    assert unit2unit(Unit.M, Unit.CM, 2.0) == 200.0
    assert unit2unit(Unit.CM, Unit.M, 200.0) == 2.0
    assert unit2unit(Unit.MM, Unit.CM, 20.0) == 2.0
    assert unit2unit(Unit.CM, Unit.MM, 2.0) == 20.0
    assert unit2unit(Unit.MM, Unit.M, 200.0) == 0.2
    assert unit2unit(Unit.M, Unit.DIP, 1) == 3779.5296
    with pytest.raises(ValueError):
        unit2unit("Not", Unit.MM, 200.0)
    with pytest.raises(ValueError):
        unit2unit(Unit.MM, "Not", 200.0)
    with pytest.raises(ValueError):
        unit2unit(Unit.MM, Unit.S, 1)
    assert unit2unit(Unit.S, Unit.MS, 1) == 1000
    assert unit2unit(Unit.MS, Unit.S, 1000) == 1
    assert unit2unit(Unit.MS, Unit.MS, 1000) == 1000
    assert unit2unit(Unit.S, Unit.S, 1) == 1


def test_conversion_matrix_unit():
    """
    Test the conversion matrix unit function.
    """
    matrix = unit2unit_matrix(Unit.M, Unit.CM)
    assert matrix[0][0] == 100.0
    assert matrix[1][1] == 100.0
    assert matrix[2][2] == 1.0
    assert (unit2unit_matrix(Unit.UNDEFINED, Unit.MM) == np.identity(3)).all()
    with pytest.raises(ValueError):
        unit2unit_matrix("Not", Unit.M)
    with pytest.raises(ValueError):
        unit2unit_matrix(Unit.M, "Not")


def test_color_conversation():
    """Validate that the UimID can be converted from s-form to byte array."""
    color: int = -1  # white color
    r, g, b, a = PathPointProperties.color(color)
    c2: int = PathPointProperties.rgba(r, g, b, a)
    assert math.isclose(color, c2, rel_tol=0.1, abs_tol=0.0)
    red: float = 0.3
    green: float = 0.3
    blue: float = 0.3
    alpha: float = 1.
    color: int = PathPointProperties.rgba(red, green, blue, alpha)
    r2, g2, b2, a2 = PathPointProperties.color(color)
    assert math.isclose(red, r2, rel_tol=0.1, abs_tol=0.0)
    assert math.isclose(green, g2, rel_tol=0.1, abs_tol=0.0)
    assert math.isclose(blue, b2, rel_tol=0.1, abs_tol=0.0)
    assert math.isclose(alpha, a2, rel_tol=0.1, abs_tol=0.0)


def test_uimid_s_form_conversation():
    """Validate that the UimID can be converted from s-form to byte array."""
    uimid: uuid.UUID = UUIDIdentifier.id_generator()
    uimid_object: UUIDIdentifier = UUIDIdentifier(uimid)
    uimid_conv: uuid.UUID = UUIDIdentifier.str_to_uimid(uimid_object.id_s_form)
    assert uimid == uimid_conv


def test_uimid_h_form_conversation():
    """Validate that the UimID can be converted from s-form to byte array."""
    uimid: uuid.UUID = UUIDIdentifier.id_generator()
    uimid_object: UUIDIdentifier = UUIDIdentifier(uimid)
    assert uimid_object.id_s_form == uimid_object.id_h_form.replace('-', '')
    uimid_conv: uuid.UUID = UUIDIdentifier.str_to_uimid(uimid_object.id_h_form)
    assert uimid == uimid_conv


def test_uuid():
    uuid_str: str = 'd40ca4a8b8987d79895e62020fbc219c'
    uuid_obj: uuid.UUID = UUIDIdentifier.str_to_uimid(uuid_str)
    id_s_form: str = Identifier.uimid_to_s_form(uuid_obj)
    assert uuid_str == id_s_form
    uuid_2_str: str = '6d586b07-87b0-4a29-b9a0-7de8d0219bae'
    uuid_2_obj: uuid.UUID = UUIDIdentifier.str_to_uimid(uuid_2_str)
    id_h_form: str = str(uuid_2_obj)
    assert uuid_2_str == id_h_form



# -*- coding: utf-8 -*-
# Copyright © 2023-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from abc import ABC
from typing import List


class Matrix4x4(ABC):
    """
    Matrix4x4
    ==========
    Collection of helpers for matrices.

    Representation of a 4x4 affine matrix

        | m00  m01  m02  m03 |
        | m10  m11  m12  m13 |
        | m20  m21  m22  m23 |
        | m30  m31  m32  m33 |

    """

    @staticmethod
    def create_scale(scale: float) -> List[List[float]]:
        """
        Create a scale matrix.
        Parameters
        ----------
        scale: float
            Scale factor

        Returns
        -------
        List[List[float]]
            4x4 scale matrix
        """
        return [[scale, 0, 0., 0.],
                [0., scale, 0., 0.],
                [0., 0, scale, 0.],
                [0., 0, 0., 1.]]

    @staticmethod
    def create_translation(translation: List[float]) -> List[List[float]]:
        """
        Create a translation matrix.
        Parameters
        ----------
        translation: List[float]
            Translation vector

        Returns
        -------
        List[List[float]]
            4x4 translation matrix

        Raises
        ------
        ValueError
            Translation must be a 3D vector
        """
        if len(translation) != 3:
            raise ValueError("Translation must be a 3D vector")
        return [[1., 0., 0., translation[0]],
                [0., 1., 0., translation[1]],
                [0., 0., 1., translation[2]],
                [0., 0., 0., 1.]]



# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from enum import Enum
from logging import Logger
from typing import List, Optional, Any

logger: Logger = logging.getLogger(__name__)

# Class tag
IS: str = '@'
# URI Templates
NODE_URI_PREFIX: str = 'uim:node/{}'
NODE_ENTITY_URI_PREFIX: str = 'uim:ne/{}'

# Document Properties
WILL_NAMESPACE: str = "will://"
DOCUMENT_NAMESPACE: str = WILL_NAMESPACE + "document/3.0/"
DOCUMENT_TITLE_OBJECT: str = DOCUMENT_NAMESPACE + 'Title'
DOCUMENT_AUTHOR_OBJECT: str = DOCUMENT_NAMESPACE + 'Author'
DOCUMENT_APPLICATION_OBJECT: str = DOCUMENT_NAMESPACE + 'Application'
DOCUMENT_APPLICATION_VERSION_OBJECT: str = DOCUMENT_NAMESPACE + 'ApplicationVersion'
DOCUMENT_LOCALE_OBJECT: str = DOCUMENT_NAMESPACE + 'Locale'

DOCUMENT_CREATION_DATE_OBJECT: str = DOCUMENT_NAMESPACE + 'CreationData'
DOCUMENT_X_MIN_PROPERTY: str = DOCUMENT_NAMESPACE + 'hasMinX'
DOCUMENT_Y_MIN_PROPERTY: str = DOCUMENT_NAMESPACE + 'hasMiny'
DOCUMENT_WIDTH_PROPERTY: str = DOCUMENT_NAMESPACE + 'Width'
DOCUMENT_HEIGHT_PROPERTY: str = DOCUMENT_NAMESPACE + 'Height'
# Device properties
WILL_NAMESPACE: str = "will://"
# Subject
INK_INPUT_DEVICE: str = WILL_NAMESPACE + "input/3.0/"
SEGMENTATION_NAMESPACE: str = WILL_NAMESPACE + "segmentation/3.0/"
DOCUMENT_NAMESPACE: str = WILL_NAMESPACE + "document/3.0/"
# Properties
DEVICE_SERIAL_NUMBER_PROPERTY: str = INK_INPUT_DEVICE + "SerialNumber"
DEVICE_MANUFACTURER_PROPERTY: str = INK_INPUT_DEVICE + "Manufacturer"
DEVICE_MODEL_PROPERTY: str = INK_INPUT_DEVICE + "Model"


class BaseSchema:
    """
    BaseSchema
    ==========
    Universal Ink Model base schema. Core structures that are used in several schemas.
    """
    STRUCTURE_VERSION: str = "1.1"
    STRUCTURES_NAMESPACE: str = f"will:uim/{STRUCTURE_VERSION}/"
    # -------------------------------------------- Classes -------------------------------------------------------------
    INK_MODEL: str = f"{STRUCTURES_NAMESPACE}InkModel"
    INK_NODE: str = f"{STRUCTURES_NAMESPACE}InkNode"
    ROOT: str = f"{STRUCTURES_NAMESPACE}Root"
    STROKE_NODE: str = f"{STRUCTURES_NAMESPACE}StrokeNode"
    STROKE_GROUP_NODE: str = f"{STRUCTURES_NAMESPACE}StrokeGroupNode"
    # -------------------------------------------- Properties ----------------------------------------------------------
    HAS_SCHEMA_INCLUDE: str = f"hasSchemaInclude"
    HAS_ROOT: str = f"hasRoot"
    HAS_CHILD: str = f"hasChild"
    HAS_PARENT: str = f"hasParent"
    HAS_REPRESENTS_VIEW: str = f"representsView"
    HAS_GENERATED_BY: str = f"generatedBy"
    HAS_PARENT_INK_MODEL: str = f"hasParentInkModel"
    HAS_STROKE: str = f"hasStroke"


class SimplifiedSegmentationSchema(BaseSchema):
    """
    SimplifiedSegmentationSchema
    ============================
    Universal Ink Model simplified content type segmentation schema.
    """
    STRUCTURE_VERSION: str = "0.1"
    STRUCTURES_NAMESPACE: str = f"will:smpl-seg/{STRUCTURE_VERSION}/"
    # -------------------------------------------- Classes -------------------------------------------------------------
    ROOT: str = f"{STRUCTURES_NAMESPACE}Root"
    SIMPLIFIED_BLOCK: str = f"{STRUCTURES_NAMESPACE}SimplifiedBlock"
    TEXT: str = f"{STRUCTURES_NAMESPACE}Text"
    MATH: str = f"{STRUCTURES_NAMESPACE}Math"
    MATH_REFERENCE: str = f"{STRUCTURES_NAMESPACE}MathReference"
    UNLABELLED: str = f"{STRUCTURES_NAMESPACE}Unlabelled"
    GARBAGE: str = f"{STRUCTURES_NAMESPACE}Garbage"
    DRAWING: str = f"{STRUCTURES_NAMESPACE}Drawing"
    DIAGRAM: str = f"{STRUCTURES_NAMESPACE}Diagram"
    ANNOTATION: str = f"{STRUCTURES_NAMESPACE}Annotation"
    # -------------------------------------------- Properties ----------------------------------------------------------
    HAS_PARENT: str = f"hasParent"
    HAS_CONTENT: str = f"hasContent"
    HAS_CONTENT_SCORE: str = f"hasContentScore"
    HAS_ALT_CONTENT: str = f"hasAltContent"
    HAS_ALT_CONTENT_SCORE: str = f"hasAltContentScore"
    HAS_LANGUAGE: str = f"hasLanguage"
    HAS_DIRECTION: str = f"hasDirection"
    HAS_CATEGORY_HINT: str = f"categoryHint"


class SegmentationSchema(BaseSchema):
    """
    SegmentationSchema
    ==================
    The segmentation schema is used to segment the content of the ink model.
    """
    SEG_STRUCTURE_VERSION: str = "0.3"
    SEG_STRUCTURES_NAMESPACE: str = f"will:seg/{SEG_STRUCTURE_VERSION}/"
    # -------------------------------------------- Classes -------------------------------------------------------------
    ROOT: str = f"{SEG_STRUCTURES_NAMESPACE}Root"
    CONTENT_BLOCK: str = f"{SEG_STRUCTURES_NAMESPACE}ContentBlock"
    TEXT_REGION: str = f"{SEG_STRUCTURES_NAMESPACE}TextRegion"
    TEXT_LINE: str = f"{SEG_STRUCTURES_NAMESPACE}TextLine"
    WORD: str = f"{SEG_STRUCTURES_NAMESPACE}Word"
    WORD_OF_STROKES: str = f"{SEG_STRUCTURES_NAMESPACE}WordOfStrokes"
    EXTENDED_TEXT_REGION: str = f"{SEG_STRUCTURES_NAMESPACE}ExtendedTextRegion"
    LINE: str = f"{SEG_STRUCTURES_NAMESPACE}Line"
    UNLABELLED_BLOCK: str = f"{SEG_STRUCTURES_NAMESPACE}UnlabelledBlock"
    UNLABELLED_ITEM_GROUP: str = f"{SEG_STRUCTURES_NAMESPACE}UnlabelledItemGroup"
    UNLABELLED_ITEM: str = f"{SEG_STRUCTURES_NAMESPACE}UnlabelledItem"
    DRAWING: str = f"{SEG_STRUCTURES_NAMESPACE}Drawing"
    DRAWING_ITEM_GROUP: str = f"{SEG_STRUCTURES_NAMESPACE}DrawingItemGroup"
    DRAWING_ITEM: str = f"{SEG_STRUCTURES_NAMESPACE}DrawingItem"
    EXPRESSION_BLOCK: str = f"{SEG_STRUCTURES_NAMESPACE}ExpressionBlock"
    GENERIC_EXPRESSION_BLOCK: str = f"{SEG_STRUCTURES_NAMESPACE}GenericExpressionBlock"
    GENERIC_EXPRESSION_ITEM_GROUP: str = f"{SEG_STRUCTURES_NAMESPACE}GenericExpressionItemGroup"
    GENERIC_EXPRESSION_ITEM: str = f"{SEG_STRUCTURES_NAMESPACE}GenericExpressionItem"
    MATH_BLOCK: str = f"{SEG_STRUCTURES_NAMESPACE}MathBlock"
    MATH_ITEM_GROUP: str = f"{SEG_STRUCTURES_NAMESPACE}MathItemGroup"
    MATH_ITEM: str = f"{SEG_STRUCTURES_NAMESPACE}MathItem"
    CHEMISTRY_BLOCK: str = f"{SEG_STRUCTURES_NAMESPACE}ChemistryBlock"
    CHEMISTRY_ITEM_GROUP: str = f"{SEG_STRUCTURES_NAMESPACE}ChemistryItemGroup"
    CHEMISTRY_ITEM: str = f"{SEG_STRUCTURES_NAMESPACE}ChemistryItem"
    PHYSICS_BLOCK: str = f"{SEG_STRUCTURES_NAMESPACE}PhysicsBlock"
    PHYSICS_ITEM_GROUP: str = f"{SEG_STRUCTURES_NAMESPACE}PhysicsItemGroup"
    PHYSICS_ITEM: str = f"{SEG_STRUCTURES_NAMESPACE}PhysicsItem"
    CONTENT_BLOCK_OF_STROKE_NODES: str = f"{SEG_STRUCTURES_NAMESPACE}ContentBlockOfStrokeNodes"
    CHEMICAL_STRUCTURE: str = f"{SEG_STRUCTURES_NAMESPACE}ChemicalStructure"
    SIGNATURE: str = f"{SEG_STRUCTURES_NAMESPACE}Signature"
    GARBAGE: str = f"{SEG_STRUCTURES_NAMESPACE}Garbage"
    UNLABELED: str = f"{SEG_STRUCTURES_NAMESPACE}Unlabeled"
    DOODLE: str = f"{SEG_STRUCTURES_NAMESPACE}Doodle"
    DIAGRAM: str = f"{SEG_STRUCTURES_NAMESPACE}Diagram"
    DIAGRAM_CONNECTOR: str = f"{SEG_STRUCTURES_NAMESPACE}DiagramConnector"
    DIAGRAM_PART: str = f"{SEG_STRUCTURES_NAMESPACE}DiagramPart"
    TABLE: str = f"{SEG_STRUCTURES_NAMESPACE}Table"
    BORDER: str = f"{SEG_STRUCTURES_NAMESPACE}Border"
    LIST: str = f"{SEG_STRUCTURES_NAMESPACE}List"
    LIST_ITEM: str = f"{SEG_STRUCTURES_NAMESPACE}ListItem"
    LIST_ITEM_BULLET: str = f"{SEG_STRUCTURES_NAMESPACE}ListItemBullet"
    ANNOTATION: str = f"{SEG_STRUCTURES_NAMESPACE}Annotation"
    MARKING: str = f"{SEG_STRUCTURES_NAMESPACE}Marking"
    CONNECTOR: str = f"{SEG_STRUCTURES_NAMESPACE}Connector"
    CORRECTION: str = f"{SEG_STRUCTURES_NAMESPACE}Correction"
    # -------------------------------------------- Properties ----------------------------------------------------------
    HAS_CHILD: str = f"hasChild"
    HAS_PARENT: str = f"hasParent"
    HAS_LANGUAGE: str = f"hasLanguage"
    HAS_CONTENT: str = f"hasContent"
    HAS_ALT_CONTENT: str = f"hasAltContent"
    HAS_CATEGORY_HINT: str = f"categoryHint"
    HAS_LATEX: str = f"hasLatex"
    HAS_MATHML: str = f"hasMathML"
    HAS_ASCII_MATH: str = f"hasASCIIMath"
    HAS_ENTITY_LABEL: str = f"hasEntityLabel"
    HAS_ENTITY_TYPE: str = f"hasEntityType"
    HAS_IS_STRUCTURAL_ENTITY: str = f"isStructuralEntity"
    HAS_IS_BASE_ENTITY: str = f"isBaseEntity"
    HAS_BULLET: str = f"bullet"
    HAS_BODY: str = f"body"
    HAS_MARKING_TYPE: str = f"markingType"
    HAS_ANNOTATES: str = f"annotates"
    HAS_CONNECTOR_TYPE: str = f"connectorType"
    HAS_CONNECTS: str = f"connects"
    HAS_CORRECTION_TYPE: str = f"correctionType"
    HAS_SOURCE: str = f"source"
    HAS_MODIFIER: str = f"modifier"


class NamedEntityRecognitionSchema(SegmentationSchema):
    """
    NamedEntityRecognitionSchema
    ============================
    The Named Entity Recognition schema is used to recognize named entities in the ink model.
    """
    NER_STRUCTURE_VERSION: str = '0.1'
    NER_STRUCTURE: str = f"will:ner/{NER_STRUCTURE_VERSION}/"
    # -------------------------------------------- Classes -------------------------------------------------------------
    NAMED_ENTITY: str = f'{NER_STRUCTURE}NamedEntity'
    # -------------------------------------------- Properties ----------------------------------------------------------
    HAS_NAMED_ENTITY: str = f"hasNamedEntity"
    HAS_IS_PART_OF_NAMED_ENTITY: str = f"isPartOfNamedEntity"
    HAS_URI: str = f"hasUri"
    HAS_IRI: str = f"hasIri"
    HAS_URN: str = f"hasUrn"
    HAS_UNIQUE_ID: str = f"hasUniqueId"
    HAS_IS_REFERENCE: str = f"isReference"
    HAS_PART: str = f"hasPart"
    HAS_CREATION_DATE: str = f"hasCreationDate"
    HAS_EXPIRATION_DATE: str = f"hasExpirationDate"
    HAS_LABEL: str = f"hasLabel"
    HAS_LANGUAGE: str = f"hasLanguage"
    HAS_PROVIDER: str = f"hasProvider"
    HAS_SOURCE: str = f"hasSource"
    HAS_CONFIDENCE: str = f"hasConfidence"
    HAS_NORM_CONFIDENCE: str = f"hasNormConfidence"
    HAS_ABSTRACT_TEXT: str = f"hasAbstractText"
    HAS_ARTICLE_URL: str = f"hasArticleUrl"
    HAS_WEBSITE_URL: str = f"hasWebsiteUrl"
    HAS_IMAGE_URL: str = f"hasImageUrl"
    HAS_THUMBNAIL_URL: str = f"hasThumbnailUrl"
    HAS_PROVIDED_ENTITY_TYPE: str = f"hasProvidedEntityType"
    HAS_PROVIDED_CATEGORY: str = f"hasProvidedCategory"
    HAS_PROVIDED_ONTOLOGY_TYPE: str = f"hasProvidedOntologyType"


class PartOfSpeechSchema(SegmentationSchema):
    """
    PosSchema
    ===================
    Part of Speech (POS) schema to tag handwritten text with part of POS tags.
    """
    POS_STRUCTURE_VERSION: str = "0.1"
    POS_STRUCTURES_NAMESPACE: str = f"will:pos/{POS_STRUCTURE_VERSION}/"
    # -------------------------------------------- Classes -------------------------------------------------------------
    PART_OF_SPEECH: str = f"{POS_STRUCTURES_NAMESPACE}PartOfSpeech"
    # -------------------------------------------- Properties ----------------------------------------------------------
    HAS_POS_ENTITY: str = f"hasPOSEntity"
    HAS_IS_PART_OF_POS_ENTITY: str = f"isPartOfPOSEntity"
    HAS_PART: str = f"hasPart"
    HAS_TAG: str = f"hasTag"


# ----------------------------------------------- Math structure -------------------------------------------------------
class MathStructureSchema(SegmentationSchema):
    """
    MathStructureSchema
    ===================
    The Math structure schema is used to represent the structure of mathematical expressions in the ink model.
    """
    MATH_STRUCTURE_VERSION: str = "0.1"
    MATH_STRUCTURES_NAMESPACE: str = f"will:math-structures/{MATH_STRUCTURE_VERSION}/"
    MATH_BLOCK_STRUCTURES: str = f"{MATH_STRUCTURES_NAMESPACE}MathBlock"
    # -------------------------------------------- Classes -------------------------------------------------------------
    MATH_BLOCK: str = f"{MATH_STRUCTURES_NAMESPACE}MathBlock"
    MATH_ITEM_GROUP: str = f"{MATH_STRUCTURES_NAMESPACE}MathItemGroup"
    MATH_ITEM: str = f"{MATH_STRUCTURES_NAMESPACE}MathItem"
    EXPRESSION_LIST: str = f"{MATH_STRUCTURES_NAMESPACE}ExpressionList"
    NUMBER: str = f"{MATH_STRUCTURES_NAMESPACE}Number"
    SYSTEM: str = f"{MATH_STRUCTURES_NAMESPACE}System"
    CASES: str = f"{MATH_STRUCTURES_NAMESPACE}Cases"
    RADICAL: str = f"{MATH_STRUCTURES_NAMESPACE}Radical"
    FENCE: str = f"{MATH_STRUCTURES_NAMESPACE}Fence"
    FRACTION: str = f"{MATH_STRUCTURES_NAMESPACE}Fraction"
    SUBSCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}Subscript"
    SUPERSCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}Superscript"
    SUB_SUPER_SCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}SubSuperScript"
    PRE_SUPER_SCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}PreSuperScript"
    PRE_SUB_SCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}PreSubScript"
    PRE_SUB_SUPER_SCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}PreSubSuperScript"
    UNDER_SCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}UnderScript"
    OVER_SCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}OverScript"
    UNDER_OVER_SCRIPT: str = f"{MATH_STRUCTURES_NAMESPACE}UnderOverScript"
    ROW: str = f"{MATH_STRUCTURES_NAMESPACE}Row"
    ROW_LIST: str = f"{MATH_STRUCTURES_NAMESPACE}RowList"
    MATRIX: str = f"{MATH_STRUCTURES_NAMESPACE}Matrix"
    GROUP: str = f"{MATH_STRUCTURES_NAMESPACE}Group"
    OPERATION: str = f"{MATH_STRUCTURES_NAMESPACE}Operation"
    RELATION: str = f"{MATH_STRUCTURES_NAMESPACE}Relation"
    SYMBOL: str = f"{MATH_STRUCTURES_NAMESPACE}Symbol"
    MATHEMATICAL_TERM: str = f"{MATH_STRUCTURES_NAMESPACE}MathematicalTerm"
    UNIT: str = f"{MATH_STRUCTURES_NAMESPACE}Unit"
    OPERATOR_SYMBOL: str = f"{MATH_STRUCTURES_NAMESPACE}OperatorSymbol"
    RELATION_SYMBOL: str = f"{MATH_STRUCTURES_NAMESPACE}RelationSymbol"
    SEPARATOR: str = f"{MATH_STRUCTURES_NAMESPACE}Separator"
    DIGIT: str = f"{MATH_STRUCTURES_NAMESPACE}Digit"
    NUMBER_ITEM: str = f"{MATH_STRUCTURES_NAMESPACE}NumberItem"
    # -------------------------------------------- Properties ----------------------------------------------------------
    HAS_CHILD: str = f"hasChild"
    HAS_ENTITY_LABEL: str = f"hasEntityLabel"
    HAS_ENTITY_TYPE: str = f"hasEntityType"
    IS_STRUCTURAL_ENTITY: str = f"isStructuralEntity"
    IS_BASE_ENTITY: str = f"isBaseEntity"
    HAS_LATEX: str = f"hasLatex"
    HAS_MATHML: str = f"hasMathML"
    HAS_ASCII_MATH: str = f"hasASCIIMath"
    HAS_EXPRESSIONS: str = f"expressions"
    HAS_OPENING_BRACKET: str = f"openingBracket"
    HAS_CLOSING_BRACKET: str = f"closingBracket"
    HAS_INDEX: str = f"index"
    HAS_RADICAL_SYMBOL: str = f"radicalSymbol"
    HAS_RADICAND: str = f"radicand"
    HAS_BODY: str = f"body"
    HAS_NUMERATOR: str = f"numerator"
    HAS_DENOMINATOR: str = f"denominator"
    HAS_FRACTION_LINE: str = f"fractionLine"
    HAS_FRACTION_TYPE: str = f"fractionType"
    HAS_SUBSCRIPT: str = f"subScript"
    HAS_SUPERSCRIPT: str = f"superScript"
    HAS_PRE_SUPERSCRIPT: str = f"preSuperScript"
    HAS_PRE_SUBSCRIPT: str = f"preSubScript"
    HAS_UNDERSCRIPT: str = f"underScript"
    HAS_OVERSCRIPT: str = f"overScript"
    HAS_ROWS: str = f"rows"
    HAS_MATRIX_TYPE: str = f"matrixType"
    HAS_OPERATION_TYPE: str = f"operationType"
    HAS_REPRESENTATION: str = f"representation"
    HAS_SYMBOL_TYPE: str = f"symbolType"
    HAS_SEPARATOR_TYPE: str = f"separatorType"


class CommonRDF:
    """
    Contains a list of used RDF types.
    """
    PRED_RDF_HAS_TYPE: str = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'  # Type according to RDF 1.1 XML Syntax.
    LOCALE: str = 'http://ogp.me/ns#locale'  # Locale as defined in Open Graph Protocol.


class CommonViews(Enum):
    """
    Contains a list of known ink model views.
    """
    CUSTOM_TREE = 'custom'  # Custom
    MAIN_INK_TREE = 'main'  # Main tree of ink strokes.
    MAIN_SENSOR_TREE = 'sdm'  # Main tree of sensor data objects.
    HWR_VIEW = 'hwr'  # Handwriting Recognition view.
    NER_VIEW = 'ner'  # Named Entity Recognition view.
    SEGMENTATION_VIEW = 'seg'  # Segmentation view.
    LEGACY_HWR_VIEW = 'will://views/3.0/HWR'  # Handwriting Recognition view (legacy constant - v3.0).
    LEGACY_NER_VIEW = 'will://views/3.0/NER'  # Named Entity Recognition view (legacy constant - v3.0).


# Mapping the view name to enum
NAME_TO_VIEW: dict = {view.value: view for view in CommonViews}


class SemanticTriple:
    """
    SemanticTriple
    ==============
    A semantic triple, or simply triple, is the atomic data entity data model.
    As its name indicates, a triple is a set of three entities that codifies a statement about semantic data in the
    form of subject predicate object expressions.

    Parameters
    ----------
    subject: str
        Subject
    predicate: str
        Predicate
    obj: str
        Object
    """

    def __init__(self, subject: str, predicate: str, obj: str):
        self.__subject = subject
        self.__predicate = predicate
        self.__object = obj

    @property
    def subject(self) -> str:
        """Subject of the statement. (`str`)"""
        return self.__subject

    @subject.setter
    def subject(self, subject: str):
        self.__subject = subject

    @property
    def predicate(self) -> str:
        """Predicate of the statement. (`str`)"""
        return self.__predicate

    @predicate.setter
    def predicate(self, predicate: str):
        self.__predicate = predicate

    @property
    def object(self) -> str:
        """Object of the statement. (`str`)"""
        return self.__object

    @object.setter
    def object(self, obj: str):
        self.__object = obj

    def __eq__(self, other: Any):
        if not isinstance(other, SemanticTriple):
            logger.warning(f'Cannot compare {type(other)} with {type(self)}')
            return False
        return (self.subject == other.subject and
                self.predicate == other.predicate and
                self.object == other.object)

    def __repr__(self):
        return (f'<Semantic triple : [subject:={self.__subject}, predicate:={self.__predicate}, '
                f'object:={self.__object}]>')

    def __dict__(self):
        return  {
            'subject': self.subject,
            'predicate': self.predicate,
            'object': self.object,
        }

    def __json__(self):
        return self.__dict__()


class TripleStore:
    """
    TripleStore
    ===========

    Encapsulates a list of triple statements.

    Parameters
    ----------
    triple_statements: List[SemanticTriple]
        List of `SemanticTriple`s
    """

    def __init__(self, triple_statements: List[SemanticTriple] = None):
        self.__triple_statement: List[SemanticTriple] = triple_statements or []

    def append(self, triple_statement: SemanticTriple):
        """Appending the triple statement.

        Parameters
        ----------
        triple_statement: SemanticTriple
            Triple that needs to be added
        """
        self.__triple_statement.append(triple_statement)

    def add_semantic_triple(self, subject: str, predicate: str, obj: str):
        """
        Adding a semantic triple.

        Parameters
        ----------
        subject: `str`
            Subject of the statement
        predicate: `str`
            Predicate of the statement
        obj: `str`
            Object of the statement
        """
        self.append(SemanticTriple(subject, predicate, obj))

    def remove_semantic_triple(self, triple: SemanticTriple):
        """
        Removes a semantic triple from list.

        Parameters
        ----------
        triple: `SemanticTriple`
            Triple to be removed
        """
        self.__triple_statement.remove(triple)

    def clear_statements(self):
        """Remove all statements."""
        self.__triple_statement = []

    @property
    def statements(self) -> List[SemanticTriple]:
        """List of triple statements. (`List[SemanticTriple]`)"""
        return self.__triple_statement

    def all_statements_for(self, subject: str, predicate: str = None) -> List[SemanticTriple]:
        """
        Returns all statements for a specific subject.

        Parameters
        ----------
        subject: `str`
            Filter for the subject URI
        predicate: `str`
            Predicate filter [optional]

        Returns
        -------
        statements: `List[SemanticTriple]`
            List of statements that match the filters.
        """
        return [s for s in self.statements if s.subject == subject and
                (predicate is None or s.predicate == predicate)]

    def filter(self, subject: Optional[str] = None, predicate: Optional[str] = None, obj: Optional[str] = None) \
            -> List[SemanticTriple]:
        """
        Returns all statements for a specific subject. The filter ares are optional, if not provided, all statements.
        Otherwise, the statements that match the filters are returned.

        Parameters
        ----------
        subject: `Optional[str]` (optional) [default: None]
            Filter for the subject URI [optional]
        predicate: `Optional[str]` (optional) [default: None]
            Predicate filter [optional]
        obj: `Optional[str]` (optional) [default: None]
            Object filter [optional]

        Returns
        -------
        statements: `List[SemanticTriple]`
            List of statements that match the filters.

        Examples
        --------
        >>> store = TripleStore()
        >>> store.add_semantic_triple('subject-1', 'predicate', 'object-1')
        >>> store.add_semantic_triple('subject-2', 'predicate', 'object-2')
        >>> store.add_semantic_triple('subject-3', 'predicate-2', 'object-3')
        >>> store.filter('subject-1', 'predicate', 'object-1')
        [Semantic triple : [subject:=subject-1, predicate:=predicate, object:=object-1]]
        >>> store.filter('predicate')
        [Semantic triple : [subject:=subject-1, predicate:=predicate, object:=object-1],
         Semantic triple : [subject:=subject-2, predicate:=predicate, object:=object-2]]
        >>> store.filter()
        [Semantic triple : [subject:=subject-1, predicate:=predicate, object:=object-1],
         Semantic triple : [subject:=subject-2, predicate:=predicate, object:=object-2],
         Semantic triple : [subject:=subject-3, predicate:=predicate-2, object:=object-3]]
        """
        results: List[SemanticTriple] = []
        for s in self.statements:
            if subject is not None and subject != s.subject:
                continue
            if predicate is not None and predicate != s.predicate:
                continue
            if obj is not None and obj != s.object:
                continue
            results.append(s)
        return results

    def determine_sem_type(self, node: 'InkNode', typedef_pred: str = CommonRDF.PRED_RDF_HAS_TYPE) -> Optional[str]:
        """
        Determine the Semantic Type of node.

        Parameters
        ----------
        node: `InkNode`
            `InkNode` to extract the semantics from
        typedef_pred: `str` (optional) [default: CommonRDF.PRED_RDF_HAS_TYPE]
            Predicate string

        Returns
        -------
        semantic_type: `str`
             Semantic type of the `InkNode`. None if the node is not found or the predicate statement.
        """
        triples: List[SemanticTriple] = self.filter(subject=node.uri, predicate=typedef_pred)
        if len(triples) == 1:
            return triples[0].object
        return None

    def __dict__(self):
        return {
            'statements': [s.__dict__() for s in self.__triple_statement]
        }

    def __iter__(self):
        return iter(self.__triple_statement)

    def __repr__(self):
        return f'<TripleStore : [statements:={self.__triple_statement}]>'



# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import functools
import logging
import uuid
from abc import ABC, abstractmethod
from typing import List, Any, Optional
from typing import Tuple
from uim.codec.parser.base import SupportedFormats
from uim.model.base import UUIDIdentifier, InkModelException
from uim.model.semantics.structures import BoundingBox
from uim.model.semantics.schema import CommonViews

logger: logging.Logger = logging.getLogger(__name__)


class URIBuilder(ABC):
    """
    URIBuilder
    ==========
    Generates URIs according to the ink model URI scheme.
    """
    SCHEME: str = "uim:"

    def __init__(self):
        pass

    @staticmethod
    def build_uri(sub_path: str, model_id: uuid.UUID = None):
        """
        Creates a URI for the specified sub-path and model identifier.
        Parameters
        ----------
        sub_path: str
            Sub-path
        model_id: UUID [default=None]
            Model identifier [optional]

        Returns
        -------
        uri: str
            URI
        """
        uri: str = URIBuilder.SCHEME
        if model_id is not None:
            uri += f'{str(model_id)}/'
        uri += f'{sub_path}/'
        return uri

    @staticmethod
    def build_named_entity_uri(uimid: uuid.UUID, model_id: uuid.UUID = None) -> str:
        """Creates a URI for the specified named entity identifier.

        Parameters
        ----------
        uimid: UUID
            UUID of the named entity
        model_id: Optional[UUID] (optional) [default=None]
            ID of the model [optional]

        Returns
        -------
        uri: str
            URI for the named entity with the ink model
        """
        if not isinstance(uimid, uuid.UUID):
            raise InkModelException("The named entity identifier should be a valid UUID.")

        return URIBuilder.build_uri('ne', model_id) + str(uimid)

    @staticmethod
    def build_entity_uri(uimid: uuid.UUID, model_id: uuid.UUID = None) -> str:
        """
        Creates a URI for the specified entity view identifier.

        Parameters
        ----------
        uimid: UUID
            UUID of the named entity
        model_id: Optional[UUID] (optional) [default=None]
            ID of the model [optional]

        Returns
        -------
        uri: str
            URI for the entity with the ink model
        """
        if not isinstance(uimid, uuid.UUID):
            raise InkModelException("The named entity identifier should be a valid UUID.")
        return URIBuilder.build_uri('', model_id) + str(uimid)

    @staticmethod
    def build_node_uri_from(node_uuid: uuid.UUID, view_name: str,
                            uri_format: SupportedFormats = SupportedFormats.UIM_VERSION_3_1_0) -> str:
        """
        Build the node URI.

        Parameters
        ----------
        node_uuid: UUID
            Node UUID
        view_name: str
            Name of the view
        uri_format: SupportedFormats (optional) [default=SupportedFormats.UIM_VERSION_3_1_0]
            URI format

        Returns
        -------
        uri: str
            URI for the node
        """
        node_uri: str = URIBuilder.build_uri('node')
        if view_name == CommonViews.LEGACY_HWR_VIEW.value:
            node_uri += 'hwr/' if uri_format == SupportedFormats.UIM_VERSION_3_1_0 else ''
        elif view_name == CommonViews.LEGACY_NER_VIEW.value:
            node_uri += 'ner/' if uri_format == SupportedFormats.UIM_VERSION_3_1_0 else ''
        else:
            node_uri += f'{view_name}/'

        node_uri += f'{node_uuid}'
        return node_uri

    @staticmethod
    def build_stroke_uri(stroke_uuid: uuid.UUID) -> str:
        """
        Build the stroke URI.

        Parameters
        ----------
        stroke_uuid: UUID
            Stroke UUID

        Returns
        -------
        uri: str
            URI for the stroke
        """
        return f"{URIBuilder.build_uri('stroke')}{str(stroke_uuid)}"

    @staticmethod
    def build_node_uri(ink_node: 'InkNode', uri_format: SupportedFormats) -> str:
        """Creates a URI for an ink node.

        Parameters
        ----------
        ink_node: InkNode
            Node in a tree
        uri_format: SupportedFormats
            URI format

        Returns
        -------
        uri: str
            URI for the node
        """
        if isinstance(ink_node, StrokeGroupNode):
            return URIBuilder.build_node_uri_from(ink_node.id, ink_node.view_name, uri_format)
        if isinstance(ink_node, StrokeNode):
            stroke_node: StrokeNode = ink_node
            node_uri: str = URIBuilder.build_node_uri_from(stroke_node.stroke.id, stroke_node.view_name, uri_format)
            if stroke_node.fragment:
                node_uri += f'#frag={stroke_node.fragment.from_point_index},{stroke_node.fragment.to_point_index}'
            return node_uri
        raise InkModelException(f"Unknown type of InkNode. Type {type(ink_node)} is not supported.")

    @staticmethod
    def build_tree_uri(tree_name: str) -> str:
        """
        Build the tree URI.
        Parameters
        ----------
        tree_name: str
            Name of the tree

        Returns
        -------
        uri: str
            URI for the tree
        """
        return f"uim:tree/{tree_name}"


class InkNode(UUIDIdentifier):
    """
    InkNode
    =======
    Represents a node in the ink tree. The node can be a `StrokeNode` or a `StrokeGroupNode`.

    The ink tree is built with a generic node structure. For building the tree the depth attribute reflects the
    depth within the tree. For serialization of the tree structure, the depth first pre-order tree serialization is
    applied.

    Each node has a unique identifier id (uri) which is relevant for the semantic statements as an identifier for the
    subject. The groupBoundingBox is optional for `StrokeGroupNode`s and assists with easier visual debugging or
    to highlight the relevant area for clickable options.

    Parameters
    ----------
    node_id: `UUID`
        Node ID as identifier.
    group_bounding_box: Optional[BoundingBox] (optional) [default=None]
        Bounding box (Group nodes only)
    """

    def __init__(self, node_id: uuid.UUID, group_bounding_box: Optional[BoundingBox] = None):
        super(UUIDIdentifier, self).__init__(node_id)
        self.__group_bounding_box: BoundingBox = group_bounding_box if group_bounding_box else BoundingBox(x=0, y=0,
                                                                                                           width=0,
                                                                                                           height=0)
        self.__parent: Optional[StrokeGroupNode] = None
        self.__tree: Optional['InkTree'] = None
        self.__transient_tag: Optional[str] = None

    @property
    def transient_tag(self) -> str:
        """Transient tag of the node. (`str`)"""
        return self.__transient_tag

    @transient_tag.setter
    def transient_tag(self, value):
        self.__transient_tag = value

    @property
    def tree(self) -> Optional['InkTree']:
        """Reference to  the respective `InkTree`. (`InkTree`, read-only)"""
        if self.__tree is None and not self.is_root():
            self.__tree = self.parent.tree
        return self.__tree

    @tree.setter
    def tree(self, value: Optional['InkTree']):
        self.__tree = value

    @property
    def root(self) -> 'InkNode':
        """Reference to the root node (`InkNode`) of the `InkTree`. (`InkNode`, read-only)"""
        self.__assert_assigned_to_a_tree__()
        return self.__tree.root

    @property
    def parent(self) -> Optional['StrokeGroupNode']:
        """Reference to the parent node (`StrokeGroupNode`) of the `InkTree`. (`InkNode`)"""
        return self.__parent

    @parent.setter
    def parent(self, value: Optional['StrokeGroupNode']):
        self.__parent = value

    def is_root(self) -> bool:
        """
        Check if the node is the root node.
        
        Returns
        -------
        flag: bool
            Flag if the `InkNode` is a root node
        """
        return self.__parent is None

    @property
    def view_name(self) -> str:
        """Name of the associated view. (`str`, read-only)"""
        self.__assert_assigned_to_a_tree__()
        return self.__tree.name

    @property
    def uri(self) -> str:
        """URI according to UIM v3.1.0 specification. (`str`, read-only)"""
        return self.__generate_uri__()

    @property
    def uri_legacy(self) -> str:
        """URI according to UIM v3.0.0 specification. (`str`, read-only)"""
        return f'uim:node/{self.id}'

    @property
    def group_bounding_box(self) -> BoundingBox:
        """Bounding box (`StrokeGroupNode`s only). (`str`)"""
        return self.__group_bounding_box

    @group_bounding_box.setter
    def group_bounding_box(self, bbox: BoundingBox):
        self.__group_bounding_box = bbox

    @abstractmethod
    def __generate_uri__(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def child_nodes_count(self) -> int:
        """
        Counts the number of child `InkNode`s.
        
        Returns
        -------
        number: int
            Number of child nodes
        """
        raise NotImplementedError()

    def is_assigned_to_a_tree(self) -> bool:
        """
        Check if the Node is assigned to a tree.
        
        Returns
        -------
        assigned: bool
            Flag if the Node is assigned to a tree
        """
        return self.tree is not None

    def __assert_not_owned__(self):
        """
        Assert that the node is not assigned to tree.
        
        Raises
        ------
        InkModelException
            If `InkNode` already assigned to a tree
        """
        if self.is_assigned_to_a_tree():
            raise InkModelException('Node already assigned to a tree.')

    def __assert_assigned_to_a_tree__(self):
        """
        Assert that the node is not assigned to a tree.
        
        Raises
        ------
        InkModelException
            If Node not assigned to a tree yet.
        """
        if not self.is_assigned_to_a_tree():
            raise InkModelException("Node not yet assigned to a tree.")

    def __dict__(self):
        return {
            'id': str(self.id),
            'uri': self.uri,
            'groupBoundingBox': self.group_bounding_box.__dict__() if self.group_bounding_box else None
        }

    def __json__(self):
        return self.__dict__()


class StrokeFragment(ABC):
    """
    StrokeFragment
    ==============
    Denotes a stroke fragment, to reference fragments of the stroke. Fragments can be used for defining semantic
    statements, for instance characters within a word, which requires to assign semantics only on a fragment of
    the stroke.
    
    Parameters
    ----------
    from_point_index: int
        The index of the first path point, which is relevant for this node.
    to_point_index: int
        The index of the last path point, which is relevant for this node.
    from_t_value: float
        The t parameter value of the first point of this node.
    to_t_value: float
        The t parameter value of the last point of this node.

    Raises
    ------
    ValueError
        Thrown if the values are negative or out of the defined range.
    """

    def __init__(self, from_point_index: int, to_point_index: int, from_t_value: float, to_t_value: float):
        if from_point_index < 0:
            raise ValueError(". The value of from_point_index must be non-negative")

        if to_point_index < from_point_index:
            raise ValueError("The value of to_point_index must be greater or equal to from_point_index")

        if from_t_value < 0. or from_t_value >= 1.:
            raise ValueError("The value of from_t_value must be in the interval [0, 1)")

        if (to_t_value <= from_t_value) or (to_t_value > 1.):
            raise ValueError("The value of to_t_value must be in the interval (fromTValue, 1]")

        self.__from_point_index: int = from_point_index
        self.__to_point_index: int = to_point_index
        self.__from_t_value: float = from_t_value
        self.__to_t_value: float = to_t_value

    @property
    def from_point_index(self) -> int:
        """Gets the index of the first path point of this node. (`int`)"""
        return self.__from_point_index

    @from_point_index.setter
    def from_point_index(self, value: int):
        self.__from_point_index = value

    @property
    def to_point_index(self) -> int:
        """Gets the index of the last path point of this node. (`int`)"""
        return self.__to_point_index

    @to_point_index.setter
    def to_point_index(self, value: int):
        self.__to_point_index = value

    @property
    def from_t_value(self) -> float:
        """Gets the t parameter value of the first point of this node. (`int`)"""
        return self.__from_t_value

    @from_t_value.setter
    def from_t_value(self, value: float):
        self.__from_t_value = value

    @property
    def to_t_value(self) -> float:
        """Gets the t parameter value of the last point of this node. (`int`)"""
        return self.__to_t_value

    @to_t_value.setter
    def to_t_value(self, value: float):
        self.__to_t_value = value

    def __eq__(self, other: Any):
        if not isinstance(other, StrokeFragment):
            logger.warning(f"Comparing StrokeFragment with {type(other)}")
            return False
        if self.from_point_index != other.from_point_index:
            logger.warning(f"StrokeFragment from_point_index is different. {self.from_point_index} != "
                           f"{other.from_point_index}")
            return False
        if self.to_point_index != other.to_point_index:
            logger.warning(f"StrokeFragment to_point_index is different. {self.to_point_index} != "
                           f"{other.to_point_index}")
            return False
        if self.from_t_value != other.from_t_value:
            logger.warning(f"StrokeFragment from_t_value is different. {self.from_t_value} != {other.from_t_value}")
            return False
        if self.to_t_value != other.to_t_value:
            logger.warning(f"StrokeFragment to_t_value is different. {self.to_t_value} != {other.to_t_value}")
            return False
        return True

    def __dict__(self):
        return {
            'from_point_index': self.from_point_index,
            'to_point_index': self.to_point_index,
            'from_t_value': self.from_t_value,
            'to_t_value': self.to_t_value
        }

    def __json__(self):
        return self.__dict__()

    def __repr__(self):
        return f'<StrokeFragment: [from_point_index:={self.from_point_index}, to_point_index:={self.to_point_index}, ' \
               f'from_t_value:={self.from_t_value}, to_t_value:={self.to_t_value}]>'


class StrokeNode(InkNode):
    """
    StrokeNode
    ==========
    Represents an ´InkNode´ that refers to a Stroke object.
    A `StrokeNode` can represent either a whole path or part of a path within the ink model.

    Parameters
    ----------
    stroke: `Stroke`
        Stroke which is referenced
    fragment: Optional[`StrokeFragment`] (optional) [default: None]
        Fragment referencing only parts of the `Stroke`

    Examples
    --------
    >>> from uim.model.ink import InkModel, InkTree
    >>> from uim.model.semantics.node import StrokeGroupNode, StrokeNode, StrokeFragment, URIBuilder
    >>> # Create the model
    >>> ink_model: InkModel = InkModel()
    >>> # Assign the group as the root of the main ink tree
    >>> ink_model.ink_tree = InkTree()
    >>> # First you need a root group to contain the strokes
    >>> root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    >>> ink_model.ink_tree.root = root
    >>>
    >>> # Add a node for stroke 0
    >>> stroke_node_0: StrokeNode = StrokeNode(stroke_0, StrokeFragment(0, 1, 0.0, 1.0))
    >>> root.add(stroke_node_0)
    >>>
    >>> # Add a node for stroke 1
    >>> root.add(StrokeNode(stroke_1, StrokeFragment(0, 1, 0.0, 1.0)))
    """

    def __init__(self, stroke: 'Stroke', fragment: Optional[StrokeFragment] = None):
        super().__init__(node_id=stroke.id)
        self.__ref_stroke: 'Stroke' = stroke
        self.__fragment: StrokeFragment = fragment

    @property
    def stroke(self) -> 'Stroke':
        """References the strokes. (`Stroke`)"""
        return self.__ref_stroke

    @stroke.setter
    def stroke(self, stroke: 'Stroke'):
        self.__ref_stroke = stroke

    @property
    def fragment(self) -> StrokeFragment:
        """`StrokeFragment` that specifies a fragment of the stroke. (`StrokeFragment`)

        Notes
        -----
        If the property value is not null, the stroke node refers to the specified fragment of the stroke.
        """
        return self.__fragment

    @fragment.setter
    def fragment(self, fragment: StrokeFragment):
        self.__fragment = fragment

    def child_nodes_count(self) -> int:
        """
        Number of child nodes.

        Returns
        -------
        num_child_nodes: int
            number of children nodes
        """
        return 0

    def __generate_uri__(self):
        return URIBuilder().build_node_uri(self, SupportedFormats.UIM_VERSION_3_1_0)

    def __dict__(self):
        return {
            'type': 'StrokeNode',  # 'type' is added to distinguish between StrokeNode and StrokeGroupNode in the JSON
            'id': str(self.id),
            'uri': self.uri,
            'stroke_id': str(self.stroke.id) if self.stroke is not None else None,
            'fragment': self.fragment.__dict__() if self.fragment else None
        }

    def __json__(self):
        return self.__dict__()

    def __eq__(self, other: Any):
        if not isinstance(other, StrokeNode):
            logger.warning(f"Comparing StrokeNode with {type(other)}")
            return False
        if self.stroke != other.stroke:
            logger.warning(f"StrokeNode is different. {self.stroke} != {other.stroke}")
            return False
        if self.fragment != other.fragment:
            logger.warning(f"StrokeNode fragment is different. {self.fragment} != {other.fragment}")
            return False
        return self.id == other.id

    def __repr__(self):
        return f'<StrokeNode: [stroke id:={self.stroke.id}]>'


class StrokeGroupNode(InkNode):
    """
    StrokeGroupNode
    ===============
    A non-leaf node, used to group ink-nodes of type `StrokeNode` and/or `StrokeGroupNode`.

    Parameters
    ----------
    uim_id: `UUID`
        Identifier of this stroke node.
    """

    def __init__(self, uim_id: uuid.UUID):
        super().__init__(node_id=uim_id)
        self.__children: List[InkNode] = []

    def add(self, node: InkNode) -> InkNode:
        """
        Adds a child node to this group.

        Parameters
        -----------
        node: `InkNode`
            The child node to be added.

        Raises
        ------
        InkModelException
            If `InkNode` already assigned to a tree or trying to add an ink node as a child, which has already a parent.
        """
        node.__assert_not_owned__()

        if node.parent is not None:
            raise InkModelException(f"Trying to add an ink node as a child, which has already a parent. Node: {node}")

        node.parent = self
        self.__children.append(node)

        if node.is_assigned_to_a_tree():
            self.tree.register_sub_tree(node)

        return node

    def remove(self, node: InkNode):
        """
        Remove child node.

        Parameters
        ----------
        node: `InkNode`
            The child node to be removed.
        """
        self.__children.remove(node)
        node.tree = None
        node.parent = None

    def child_nodes_count(self) -> int:
        """
        Number of child nodes.

        Returns
        -------
        number_of_child_nodes: `int`
            Number of child nodes
        """
        return len(self.__children)

    def child_group_nodes_count(self) -> int:
        """
        Number of child group nodes.

        Returns
        -------
        number_of_child_nodes: `int`
            Number of child group nodes
        """
        count: int = 0
        if self.child_nodes_count() > 0:
            for n in self.__children:
                count = count + 1 if isinstance(n, StrokeGroupNode) else 0
        return count

    def child_stroke_nodes_count(self) -> int:
        """
        Number of child stroke nodes.

        Returns
        -------
        number_of_child_nodes: `int`
            Number of child stroke group nodes
        """
        count: int = 0
        if self.child_nodes_count() > 0:
            for n in self.__children:
                count += 1 if isinstance(n, StrokeNode) else 0
        return count

    def sort_children(self, lambda_sort_func: Any, reverse: bool = False):
        """
        Sorts the children based on a sorting function.

        Parameters
        ----------
        lambda_sort_func: Any
            Sorting function for children.
        reverse: bool (optional) [default:=False]
            Flag for reversed order [default:=False]
        """
        self.__children = sorted(self.__children, key=functools.cmp_to_key(lambda_sort_func), reverse=reverse)

    @property
    def children(self) -> Tuple[InkNode]:
        """Children of the node. (`Tuple[InkNode]`, read-only)"""
        return tuple(self.__children)

    def __generate_uri__(self, uri_format: SupportedFormats = SupportedFormats.UIM_VERSION_3_1_0) -> str:
        return URIBuilder().build_node_uri(self, uri_format)

    def __dict__(self):
        return {
            'type': 'StrokeGroupNode',
            # 'type' is added to distinguish between StrokeNode and StrokeGroupNode in the JSON
            'id': str(self.id),
            'uri': self.uri,
            'children': [child.__dict__() for child in self.__children]
        }

    def __json__(self):
        return self.__dict__()

    def __eq__(self, other: Any):
        if not isinstance(other, StrokeGroupNode):
            logger.warning(f"Comparing StrokeGroupNode with {type(other)}")
            return False
        return self.id == other.id

    def __repr__(self):
        return f'<StrokeGroupNode: [uri:={self.uri}, children:={len(self.children)}]>'



# -*- coding: utf-8 -*-
"""
Semantics
=========
The semantics package contains the following modules:
- node: The node module contains the base classes for the Universal Ink Model (UIM) nodes.
- structures: The structures module contains the data structures used within the Universal Ink Model (UIM).
- schema: The schema module contains the schema classes for the Universal Ink Model (UIM).
"""

__all__ = ['node', 'structures', 'schema']



# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from abc import ABC
from enum import Enum
from math import isclose
from typing import List, Tuple, Any, Optional, Union

import uim.codec.format.UIM_3_0_0_pb2 as uim
from uim.model.base import InkModelException

logger: logging.Logger = logging.getLogger(__name__)
TOLERANCE_VALUE_COMPARISON: float = 1e-2


class RotationMode(Enum):
    """
    RotationMode
    ============
    Brush rotation modes.
    """
    NONE = 0
    """Indicates that the shape will not be rotated."""
    RANDOM = 1
    """Indicates that the shape will be rotated randomly before it renders."""
    TRAJECTORY = 2
    """Indicates that the shape will be rotated to match the path trajectory."""


class BlendMode(Enum):
    """
    BlendMode
    =========
    The blend mode enum lists the different blend modes which can be applied to raster brushes.
    """

    SOURCE_OVER = uim.SOURCE_OVER
    """This is the default setting and draws new shapes on top of the existing canvas content. Also known as NORMAL."""
    DESTINATION_OVER = uim.DESTINATION_OVER
    """New shapes are drawn behind the existing canvas content. Also known as NORMAL_REVERSE."""
    DESTINATION_OUT = uim.DESTINATION_OUT
    """The existing content is kept where it doesn't overlap the new shape. Also known as ERASE."""
    LIGHTER = uim.LIGHTER
    """Where both shapes overlap the color is determined by adding color values. Also known as ADD."""
    COPY = uim.COPY
    """Only the new shape is shown. Also known as NONE."""
    MIN = uim.MIN
    """The result is the minimum of both color. The result is a darker color."""
    MAX = uim.MAX
    """The result is the maximum of both color. The result is a lighter color."""


class BlendModeURIs(ABC):
    """
    BlendModeURIs
    =============
    URIs for the different blend modes.
    """
    SOURCE_OVER: str = "will://rasterization/3.0/blend-mode/SourceOver"
    """URI for BlendMode.SourceOver"""
    DESTINATION_OVER: str = "will://rasterization/3.0/blend-mode/DestinationOver"
    """URI for BlendMode.DestinationOver"""
    DESTINATION_OUT: str = "will://rasterization/3.0/blend-mode/DestinationOut"
    """URI for BlendMode.DestinationOut"""
    LIGHTER: str = "will://rasterization/3.0/blend-mode/Lighter"
    """URI for BlendMode.Lighter"""
    COPY: str = "will://rasterization/3.0/blend-mode/Copy"
    """URI for BlendMode.Copy"""
    MIN: str = "will://rasterization/3.0/blend-mode/Min"
    """URI for BlendMode.Min"""
    MAX: str = "will://rasterization/3.0/blend-mode/Max"
    """URI for BlendMode.Max"""


class BrushPolygonUri(ABC):
    """
    BrushPolygonUri
    ===============
    Represents a vector brush shape that is specified with a URI.

    Parameters
    ----------
    uri: ´str´
        URI string that identifies the shape.
    min_scale: ´float´
        Minimum scale of the brush sample, after which this shape is applied.
    """

    def __init__(self, uri: str, min_scale: float):
        self.__uri: str = uri
        self.__min_scale: float = min_scale

    @property
    def shape_uri(self) -> str:
        """
        URI string that identifies the shape. (´str´)
        """
        return self.__uri

    @property
    def min_scale(self) -> float:
        """
        Minimum scale of the brush sample, after which this shape is applied. (´float´)

        Notes
        -----
        This value is used by the brush applier in order to pick the proper shape according to the actual brush scale.
        """
        return self.__min_scale

    def __dict__(self):
        return {
            "uri": self.shape_uri,
            "min_scale": self.min_scale
        }

    def __json__(self):
        return self.__dict__()

    def __eq__(self, other: Any):
        if not isinstance(other, BrushPolygonUri):
            logger.warning(f'Cannot compare BrushPolygonUri with {type(other)}')
            return False
        return self.shape_uri == other.shape_uri and isclose(self.min_scale, other.min_scale,
                                                             abs_tol=TOLERANCE_VALUE_COMPARISON)

    def __repr__(self):
        return f'<VectorBrushPrototypeUri : [uri:={self.__uri}, min scale:={self.min_scale}]>'


class BrushPolygon(ABC):
    """
    BrushPolygon
    ============
    Describes vector brush prototype.

    Parameters
    ----------
    points: Optional[List[Tuple[float, float]]] (optional) [default: None]
        List of points for polygon
    indices: Optional[List[int]] (optional) [default: None]
        List of indexes
    """

    def __init__(self, min_scale: float, points: Optional[List[Tuple[float, float]]] = None,
                 indices: Optional[List[int]] = None):
        self.__min_scale: float = min_scale
        self.__points: List[Tuple[float, float]] = points or []
        self.__indices: List[int] = indices or []

    @property
    def points(self) -> List[Tuple[float, float]]:
        """List of coordinates for x value. (`List[Tuple[float, float]]`, read-only)"""
        return self.__points

    @property
    def coord_x(self) -> List[float]:
        """List of coordinates for x value. (`List[float]`, read-only)"""
        if len(self.__points) == 0:
            return []
        if isinstance(self.__points[0], tuple) and len(self.__points[0]) > 0:
            return [p[0] for p in self.points]
        raise InkModelException("The points do not have x value.")

    @property
    def coord_y(self) -> List[float]:
        """List of coordinates for y value. (`List[float]`, read-only)"""
        if len(self.__points) == 0:
            return []
        if isinstance(self.__points[0], tuple) and len(self.__points[0]) > 1:
            return [p[1] for p in self.points]
        raise InkModelException("The points do not have y value.")

    @property
    def coord_z(self) -> List[float]:
        """List of coordinates for z value. (`List[float]`, read-only)"""
        if len(self.__points) == 0:
            return []
        if isinstance(self.__points[0], tuple) and len(self.__points[0]) > 2:
            return [p[2] for p in self.points]
        return []

    @property
    def indices(self) -> List[int]:
        """List of indices of brush prototype [for 3D rendering]. (`List[int]`, read-only)"""
        return self.__indices

    @property
    def min_scale(self) -> float:
        """
        Minimum scale of the brush sample, after which this shape is applied. (`float`, read-only)

        Notes
        -----
        This value is used by the brush applier in order to pick the proper shape according to the actual brush scale.
        """
        return self.__min_scale

    def __dict__(self):
        return {
            "indices": self.indices,
            "min_scale": self.min_scale,
            "coord_x": self.coord_x,
            "coord_y": self.coord_y,
            "coord_z": self.coord_z
        }

    def __json__(self):
        return self.__dict__()

    def __eq__(self, other: Any):
        if not isinstance(other, BrushPolygon):
            logger.warning(f'Cannot compare BrushPolygon with {type(other)}')
            return False
        if self.__min_scale != other.min_scale:
            logger.warning(f'BrushPolygon min scale mismatch: {self.min_scale} != {other.min_scale}')
            return False
        if self.__indices != other.indices:
            logger.warning(f'BrushPolygon indices mismatch: {self.indices} != {other.indices}')
            return False
        for p1, p2 in zip(self.points, other.points):
            if not (isclose(p1[0], p2[0], abs_tol=TOLERANCE_VALUE_COMPARISON) and
                    isclose(p1[1], p2[1], abs_tol=TOLERANCE_VALUE_COMPARISON)):
                logger.warning(f'BrushPolygon point mismatch: {p1} != {p2}')
                return False
        return True

    def __repr__(self):
        return f'<VectorBrushPrototype : [#points:={len(self.__points)}]>'


class Brush(ABC):
    """
    Brush
    -----
    Abstract class for brushes.
    """


class VectorBrush(Brush):
    """
    VectorBrush
    ===========
    A configuration which allows rendering of an interpolated Catmull-Rom spline as a vector spline by applying a
    specific polygon for each interpolated point, depending on its size and merging result afterwards.

    Parameters
    ----------
    name: str
        Name of the brush
    prototypes: list
        Prototypes for brush
    spacing: float (optional) [default: 1.]
        Spacing value
    """

    def __init__(self, name: str, prototypes: List[Any] = None, spacing: float = 1.):
        self.__name: str = name
        self.__prototypes: List[Any] = prototypes or []
        self.__spacing: float = spacing

    @property
    def name(self) -> str:
        """Name of the vector brush. (`str`, read-only)"""
        return self.__name

    @property
    def prototypes(self) -> List[Union[BrushPolygon, BrushPolygonUri]]:
        """ Polygon prototype for the brush. (`List[Any]`,  read-only).

        Notes
        -----
        The prototype can be either URI-based for prototype with a defined polygon geometry.

        See also
        --------
        - `BrushPolygon` - List of polygon sample points
        - `BrushPolygonUri` - List of URI polygon prototype
        """
        return self.__prototypes

    @property
    def spacing(self) -> float:
        """Spacing value. (`float`, read-only)"""
        return self.__spacing

    def __dict__(self):
        return {
            'name': self.name,
            'prototypes': [p.__dict__() for p in self.prototypes],
            'spacing': self.spacing
        }

    def __json__(self):
        return self.__dict__()

    def __eq__(self, other: Any):
        if not isinstance(other, VectorBrush):
            logger.warning(f'Cannot compare VectorBrush with {type(other)}')
            return False
        return self.name == other.name and self.prototypes == other.prototypes and self.spacing == other.spacing

    def __repr__(self):
        return f'<VectorBrush : [name:={self.__name}]>'


class RasterBrush(Brush):
    """
    RasterBrush
    ===========
    A configuration which allows rendering of an interpolated Catmull-Rom spline as a raster image by applying a
    specific sprite for each interpolated point, depending on its size.

    Parameters
    ----------
    name: `str`
        Brush descriptor
    spacing: `float`
        Distance between neighbour particles.
    scattering: `float`
        The scattering along the curve normal.
    rotation: `RotationMode`
        The particle rotation mode of the brush.
    shape_textures: Optional[List[bytes]] (optional) [default: None]
        List of png images that contains the shape texture.
    shape_texture_uris: Optional[List[str]] (optional) [default: None]
        List of URIs that contains the shape texture.
    fill_texture: Optional[bytes]
        List of png image that contains the fill texture.
    fill_texture_uri: Optional[str] (optional) [default: None]
        List of URIs that describes the fill textures.
    fill_width: `float` (optional) [default: 0.]
        Width of the fill tile.
    fill_height: `float` (optional) [default: 0.]
        Height of the fill tile.
    randomize_fill: `bool` (optional) [default: False]
        Specifies whether the fill texture is randomly displaced.
    blend_mode: `BlendMode` (optional) [default: BlendMode.SOURCE_OVER]
        The applied blend mode.
    """

    def __init__(self, name: str, spacing: float, scattering: float, rotation: RotationMode,
                 shape_textures: Optional[List[bytes]] = None, shape_texture_uris: Optional[List[str]] = None,
                 fill_texture: Optional[bytes] = None, fill_texture_uri: Optional[str] = None,
                 fill_width: float = 0., fill_height: float = 0.,
                 randomize_fill: bool = False, blend_mode: BlendMode = BlendMode.SOURCE_OVER):
        self.__name: str = name
        self.__spacing: float = spacing
        self.__scattering: float = scattering
        self.__rotation: RotationMode = rotation
        self.__shape_textures: List[bytes] = shape_textures if shape_textures is not None else []
        self.__shape_texture_uri: List[str] = shape_texture_uris if shape_texture_uris is not None else []
        self.__fill_texture: bytes = fill_texture if fill_texture is not None else b''
        self.__fill_texture_uri: str = fill_texture_uri if fill_texture_uri is not None else ''
        self.__fill_width: float = fill_width
        self.__fill_height: float = fill_height
        self.__randomize_fill: bool = randomize_fill
        self.__blend_mode: BlendMode = blend_mode

    @property
    def name(self) -> str:
        """ Name of the raster brush. (`str`, read-only)"""
        return self.__name

    @property
    def spacing(self) -> float:
        """Distance between neighbour particles. (`float`, read-only)"""
        return self.__spacing

    @property
    def scattering(self) -> float:
        """The scattering along the curve normal. (`float`, read-only)"""
        return self.__scattering

    @property
    def rotation(self) -> RotationMode:
        """The particle rotation mode of the brush. (`float`, read-only)"""
        return self.__rotation

    @property
    def shape_textures(self) -> List[bytes]:
        """List of png images that contains the shape texture; byte array with PNG content.
        (`List[bytes]`, read-only)"""
        return self.__shape_textures

    @property
    def shape_texture_uris(self) -> List[str]:
        """URI identifying the fillTexture. (`List[str]`, read-only)"""
        return self.__shape_texture_uri

    @property
    def fill_texture(self) -> bytes:
        """List of png image that contains the fill texture. (`bytes`, read-only)"""
        return self.__fill_texture

    @property
    def fill_texture_uri(self) -> str:
        """The fill texture URI. (`str`, read-only)"""
        return self.__fill_texture_uri

    @property
    def fill_width(self) -> float:
        """Width of the fill tile. (`float`, read-only)"""
        return self.__fill_width

    @property
    def fill_height(self) -> float:
        """Height of the fill tile. (`float`, read-only)"""
        return self.__fill_height

    @property
    def randomize_fill(self) -> bool:
        """Specifies whether the fill texture is randomly displaced. (`bool`, read-only)"""
        return self.__randomize_fill

    @property
    def blend_mode(self) -> BlendMode:
        """The applied blend mode. (`BlendMode`, read-only)"""
        return self.__blend_mode

    def __dict__(self):
        return {
            'name': self.name,
            'spacing': self.spacing,
            'scattering': self.scattering,
            'rotation': self.rotation,
            'shape_texture_uris': self.shape_texture_uris,
            'fill_texture_uri': self.fill_texture_uri,
            'fill_width': self.fill_width,
            'fill_height': self.fill_height,
            'randomize_fill': self.randomize_fill,
            'blend_mode': self.blend_mode
        }

    def __json__(self):
        return self.__dict__()

    def __eq__(self, other: Any):
        if not isinstance(other, RasterBrush):
            logger.warning(f'Cannot compare RasterBrush with {type(other)}')
            return False
        if self.name != other.name:
            logger.warning(f'RasterBrush name mismatch: {self.__name} != {other.name}')
            return False
        if not isclose(self.spacing, other.spacing, abs_tol=TOLERANCE_VALUE_COMPARISON):
            logger.warning(f'RasterBrush spacing mismatch: {self.spacing} != {other.spacing}')
            return False
        if not isclose(self.scattering, other.scattering, abs_tol=TOLERANCE_VALUE_COMPARISON):
            logger.warning(f'RasterBrush scattering mismatch: {self.scattering} != {other.scattering}')
            return False
        if self.rotation != other.rotation:
            logger.warning(f'RasterBrush rotation mismatch: {self.rotation} != {other.rotation}')
            return False
        if self.shape_textures != other.shape_textures:
            logger.warning(f'RasterBrush shape textures mismatch: {self.shape_textures} != {other.shape_textures}')
            return False
        if self.shape_texture_uris != other.shape_texture_uris:
            logger.warning(f'RasterBrush shape texture URIs mismatch: {self.shape_texture_uris} != '
                           f'{other.shape_texture_uris}')
            return False
        if self.fill_texture != other.fill_texture:
            logger.warning(f'RasterBrush fill texture mismatch: {self.fill_texture} != {other.fill_texture}')
            return False
        if self.fill_texture_uri != other.fill_texture_uri:
            logger.warning(f'RasterBrush fill texture URI mismatch: {self.fill_texture_uri} != '
                           f'{other.fill_texture_uri}')
            return False
        if not isclose(self.fill_width, other.fill_width, abs_tol=TOLERANCE_VALUE_COMPARISON):
            logger.warning(f'RasterBrush fill width mismatch: {self.fill_width} != {other.fill_width}')
            return False
        if not isclose(self.fill_height, other.fill_height, abs_tol=TOLERANCE_VALUE_COMPARISON):
            logger.warning(f'RasterBrush fill height mismatch: {self.fill_height} != {other.fill_height}')
            return False
        if self.randomize_fill != other.randomize_fill:
            logger.warning(f'RasterBrush randomize fill mismatch: {self.randomize_fill} != {other.randomize_fill}')
            return False
        if self.blend_mode != other.blend_mode:
            logger.warning(f'RasterBrush blend mode mismatch: {self.blend_mode} != {other.blend_mode}')
            return False
        return True

    def __repr__(self):
        return f'<RasterBrush : [name:={self.name}]>'


class Brushes:
    """
    Brushes
    ========
    Brush descriptions, needed for uim rasterization.

    Parameters
    ----------
    vector_brushes: Optional[List[VectorBrush]] (optional) [default: None]
        List of vector brushes
    raster_brushes: Optional[List[RasterBrush] (optional) [default: None]
        List of raster brushes
    """

    def __init__(self, vector_brushes: Optional[List[VectorBrush]] = None,
                 raster_brushes: Optional[List[RasterBrush]] = None):
        self.__vector_brushes: List[VectorBrush] = vector_brushes or []
        self.__raster_brushes: List[RasterBrush] = raster_brushes or []

    @property
    def vector_brushes(self) -> List[VectorBrush]:
        """
        List of defined vector brushes. (`List[VectorBrush]`)"""
        return self.__vector_brushes

    @property
    def raster_brushes(self) -> List[RasterBrush]:
        """
        List of defined raster brushes. (`List[RasterBrush]`)
        """
        return self.__raster_brushes

    def add_vector_brush(self, brush: VectorBrush):
        """
        Adding a vector brush.

        Parameters
        ----------
        brush: `VectorBrush`
            Vector brush
        """
        self.vector_brushes.append(brush)

    def add_raster_brush(self, brush: RasterBrush):
        """
        Adding raster brush.

        Parameters
        ----------
        brush: `RasterBrush`
            Raster brush
        """
        self.raster_brushes.append(brush)

    def remove_vector_brush(self, name: str):
        """
        Remove vector brush from brushes.
        If the brush with the name does not exist, no operation is performed.

        Parameters
        ----------
        name: `str`
            Name of brush that should be remove.
        """
        for v_i, _ in enumerate(self.__vector_brushes):
            if self.__vector_brushes[v_i].name == name:
                del self.__vector_brushes[v_i]
                break

    def remove_raster_brush(self, name: str):
        """
        Remove raster brush from brushes. If the brush with the name does not exist, no operation is performed.

        Parameter
        ---------
        name: str
            Name of brush that should be remove
        """
        for v_i, _ in enumerate(self.__raster_brushes):
            if self.__raster_brushes[v_i].name == name:
                del self.__raster_brushes[v_i]
                break

    def __dict__(self):
        return {
            'vector_brushes': [b.__dict__() for b in self.vector_brushes],
            'raster_brushes': [b.__dict__() for b in self.raster_brushes]
        }

    def __json__(self):
        return self.__dict__()

    def __eq__(self, other: Any):
        if not isinstance(other, Brushes):
            logger.warning(f'Cannot compare Brushes with {type(other)}')
            return False
        if len(self.vector_brushes) != len(other.vector_brushes):
            logger.warning(f'VectorBrushes length mismatch: {len(self.vector_brushes)} != {len(other.vector_brushes)}')
            return False
        for v1, v2 in zip(self.vector_brushes, other.vector_brushes):
            if v1 != v2:
                logger.warning(f'VectorBrush mismatch: {v1} != {v2}')
                return False
        if len(self.raster_brushes) != len(other.raster_brushes):
            logger.warning(f'RasterBrushes length mismatch: {len(self.raster_brushes)} != {len(other.raster_brushes)}')
            return False
        for r1, r2 in zip(self.raster_brushes, other.raster_brushes):
            if r1 != r2:
                logger.warning(f'RasterBrush mismatch: {r1} != {r2}')
                return False
        return True

    def __repr__(self):
        return f'<Brushes : [raster brush:=#{len(self.__raster_brushes)}, vector brush:=#{len(self.__vector_brushes)}]>'
